/*	$NetBSD$	*/

/*
 * Copyright (c) 2000 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Kazuki Sakamoto <sakamoto@netbsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * Yamaha YMF724[B-F]/740[B-C]/744/754
 *
 * Documentation links:
 * 
 * ftp://ftp.alsa-project.org/pub/manuals/yamaha/
 * ftp://ftp.alsa-project.org/pub/manuals/yamaha/pci/
 */

#include "mpu.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/fcntl.h>
#include <sys/malloc.h>
#include <sys/device.h>
#include <sys/proc.h>

#include <dev/pci/pcidevs.h>
#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>

#include <sys/audioio.h>
#include <dev/audio_if.h>
#include <dev/mulaw.h>
#include <dev/auconv.h>
#include <dev/ic/ac97reg.h>
#include <dev/ic/ac97var.h>
#include <dev/ic/mpuvar.h>

#include <uvm/uvm_extern.h>

#include <machine/bus.h>
#include <machine/intr.h>

#include <dev/microcode/ymf/ymf_hwmcode.h>
#include <dev/pci/ymfreg.h>
#include <dev/pci/ymfvar.h>

/* Debug */
#ifdef AUDIO_DEBUG
#define DPRINTF(x)	if (ymfdebug) printf x
#define DPRINTFN(n,x)	if (ymfdebug>(n)) printf x
int	ymfdebug = 0;
#else
#define DPRINTF(x)
#define DPRINTFN(n,x)
#endif

int	ymf_match __P((struct device *, struct cfdata *, void *));
void	ymf_attach __P((struct device *, struct device *, void *));
int	ymf_intr __P((void *));

#define DMAADDR(p) ((p)->map->dm_segs[0].ds_addr)
#define KERNADDR(p) ((void *)((p)->addr))

int	ymf_allocmem __P((struct ymf_softc *, size_t, size_t, struct ymf_dma *));
int	ymf_freemem __P((struct ymf_softc *, struct ymf_dma *));

#ifndef AUDIO_DEBUG
#define YWRITE1(sc, r, x) bus_space_write_1((sc)->memt, (sc)->memh, (r), (x))
#define YWRITE2(sc, r, x) bus_space_write_2((sc)->memt, (sc)->memh, (r), (x))
#define YWRITE4(sc, r, x) bus_space_write_4((sc)->memt, (sc)->memh, (r), (x))
#define YREAD1(sc, r) bus_space_read_1((sc)->memt, (sc)->memh, (r))
#define YREAD2(sc, r) bus_space_read_2((sc)->memt, (sc)->memh, (r))
#define YREAD4(sc, r) bus_space_read_4((sc)->memt, (sc)->memh, (r))
#else

u_int16_t YREAD2(struct ymf_softc *sc,bus_size_t r);
u_int32_t YREAD4(struct ymf_softc *sc,bus_size_t r);
void YWRITE1(struct ymf_softc *sc,bus_size_t r,u_int8_t x);
void YWRITE2(struct ymf_softc *sc,bus_size_t r,u_int16_t x);
void YWRITE4(struct ymf_softc *sc,bus_size_t r,u_int32_t x);

u_int16_t YREAD2(struct ymf_softc *sc,bus_size_t r)
{
  DPRINTFN(5, (" YREAD2(0x%lX)\n",(unsigned long)r));
  return bus_space_read_2(sc->memt,sc->memh,r);
}
u_int32_t YREAD4(struct ymf_softc *sc,bus_size_t r)
{
  DPRINTFN(5, (" YREAD4(0x%lX)\n",(unsigned long)r));
  return bus_space_read_4(sc->memt,sc->memh,r);
}
void YWRITE1(struct ymf_softc *sc,bus_size_t r,u_int8_t x)
{
  DPRINTFN(5, (" YWRITE1(0x%lX,0x%lX)\n",(unsigned long)r,(unsigned long)x));
  bus_space_write_1(sc->memt,sc->memh,r,x);
}
void YWRITE2(struct ymf_softc *sc,bus_size_t r,u_int16_t x)
{
  DPRINTFN(5, (" YWRITE2(0x%lX,0x%lX)\n",(unsigned long)r,(unsigned long)x));
  bus_space_write_2(sc->memt,sc->memh,r,x);
}
void YWRITE4(struct ymf_softc *sc,bus_size_t r,u_int32_t x)
{
  DPRINTFN(5, (" YWRITE4(0x%lX,0x%lX)\n",(unsigned long)r,(unsigned long)x));
  bus_space_write_4(sc->memt,sc->memh,r,x);
}
#endif

#define	YWRITEREGION4(sc, r, x, c)	\
	bus_space_write_region_4((sc)->memt, (sc)->memh, (r), (x), (c) / 4)

struct cfattach ymf_ca = {
	sizeof(struct ymf_softc), ymf_match, ymf_attach
};

int	ymf_open __P((void *, int));
void	ymf_close __P((void *));
int	ymf_query_encoding __P((void *, struct audio_encoding *));
int	ymf_set_params __P((void *, int, int, struct audio_params *, struct audio_params *));
int	ymf_round_blocksize __P((void *, int));
int	ymf_trigger_output __P((void *, void *, void *, int, void (*)(void *),
	    void *, struct audio_params *));
int	ymf_trigger_input __P((void *, void *, void *, int, void (*)(void *),
	    void *, struct audio_params *));
int	ymf_halt_output __P((void *));
int	ymf_halt_input __P((void *));
int	ymf_getdev __P((void *, struct audio_device *));
int	ymf_mixer_set_port __P((void *, mixer_ctrl_t *));
int	ymf_mixer_get_port __P((void *, mixer_ctrl_t *));
void   *ymf_malloc __P((void *, int, size_t, int, int));
void	ymf_free __P((void *, void *, int));
size_t	ymf_round_buffersize __P((void *, int, size_t));
paddr_t ymf_mappage __P((void *, void *, off_t, int));
int	ymf_get_props __P((void *));
int	ymf_query_devinfo __P((void *addr, mixer_devinfo_t *dip));

int     ymf_attach_codec __P((void *sc, struct ac97_codec_if *));
int	ymf_read_codec __P((void *sc, u_int8_t a, u_int16_t *d));
int	ymf_write_codec __P((void *sc, u_int8_t a, u_int16_t d));
void    ymf_reset_codec __P((void *sc));
int     ymf_get_portnum_by_name __P((struct ymf_softc *, char *, char *,
				 char *));

static int ymf_get_dstype __P((int));
static int ymf_download_mcode __P((struct ymf_softc *));
static int ymf_allocate_slots __P((struct ymf_softc *));
static void ymf_configure_legacy __P((struct device *arg));
static void ymf_enable_dsp __P((struct ymf_softc *));
static int ymf_disable_dsp __P((struct ymf_softc *));
static int ymf_ready_codec __P((struct ymf_codec_softc *));
static int ymf_halt __P((struct ymf_softc *));
static u_int32_t ymf_get_lpfq __P((u_int));
static u_int32_t ymf_get_lpfk __P((u_int));
static struct ymf_dma *ymf_find_dma __P((struct ymf_softc *, void *));

#ifdef AUDIO_DEBUG
static void ymf_dump_play_slot __P((struct ymf_softc *, int));
#define	YMF_DUMP_PLAY_SLOT(n,sc,bank) \
	if (ymfdebug > (n)) ymf_dump_play_slot(sc, bank)
#else
#define	YMF_DUMP_PLAY_SLOT(n,sc,bank)
#endif /* AUDIO_DEBUG */

static struct audio_hw_if ymf_hw_if = {
	ymf_open,
	ymf_close,
	NULL,
	ymf_query_encoding,
	ymf_set_params,
	ymf_round_blocksize,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	ymf_halt_output,
	ymf_halt_input,
	NULL,
	ymf_getdev,
	NULL,
	ymf_mixer_set_port,
	ymf_mixer_get_port,
	ymf_query_devinfo,
	ymf_malloc,
	ymf_free,
	ymf_round_buffersize,
	ymf_mappage,
	ymf_get_props,
	ymf_trigger_output,
	ymf_trigger_input,
};

struct audio_device ymf_device = {
	"Yamaha AudioPCI",
	"",
	"ymf"
};

#ifdef AUDIO_DEBUG
static void
ymf_dump_play_slot(sc, bank)
	struct ymf_softc *sc;
	int bank;
{
	int i, j;
	u_int32_t *p;
	u_int32_t num;
	struct ymf_dma *dma;

	for (i = 0; i < 64; i++) {
		printf("pbank[%d] = %p,", i*2, sc->pbank[i*2]);
		printf("pbank[%d] = %p\n", i*2+1, sc->pbank[i*2+1]);
	}

	p = (u_int32_t*)sc->pbase;
	for (i=0;i<65;i++) {
	  printf("pbase + %d:0x%x\n",i,*p);
	  p++;
	}

	num = *(u_int32_t*)sc->pbase;
	printf("num = %d\n",num);

	for (i = 0; i < num; i++) {

		p = (u_int32_t *)sc->pbank[i];

		dma = ymf_find_dma(sc,(void*)p);

		printf("  pbank[%d] : %p(%p)\n",i,p,(void*)vtophys((vaddr_t)p));
		for (j = 0; j < sizeof(struct play_slot_ctrl_bank) /
		    sizeof(u_int32_t); j++) {
			printf("    0x%02x: 0x%08x\n",
				j * sizeof(u_int32_t), *p++);
		}
		/*
		p = (u_int32_t *)sc->pbank[i*2+1];
		printf("  pbank[%d] : %p\n",i*2+1,p);
		for (j = 0; j < sizeof(struct play_slot_ctrl_bank) /
		    sizeof(u_int32_t); j++) {
			printf("    0x%02x: 0x%08x\n",
				j * sizeof(u_int32_t), *p++);
				delay(1);
		}	
		*/
	}
}
#endif /* AUDIO_DEBUG */

static int
ymf_get_dstype(id)
	int id;
{

	switch (PCI_PRODUCT(id)) {
	case PCI_PRODUCT_YAMAHA_YMF724:
	case PCI_PRODUCT_YAMAHA_YMF740:
		return YMF_DS_1;

	case PCI_PRODUCT_YAMAHA_YMF740C:
	case PCI_PRODUCT_YAMAHA_YMF724F:
	case PCI_PRODUCT_YAMAHA_YMF744B:
	case PCI_PRODUCT_YAMAHA_YMF754:
		return YMF_DS_1E;
	}

	return -1;
}

static int
ymf_download_mcode(sc)
	struct ymf_softc *sc;
{
	u_int ctrl;
	const u_int32_t *p;
	size_t size;
	int dstype;

	static struct {
		const u_int32_t *mcode;
		size_t size;
	} ctrls[] = {
		{ymf_ds1_ctrl_mcode, sizeof(ymf_ds1_ctrl_mcode)},
		{ymf_ds1e_ctrl_mcode, sizeof(ymf_ds1e_ctrl_mcode)},
	};

	dstype = ymf_get_dstype(sc->sc_id);
	if (dstype < 0)
		return 1;

	YWRITE4(sc, YMF_DAC_OUT_VOLUME, 0);
	if (ymf_disable_dsp(sc))
		return 1;

	/* DS-XG software reset */
        YWRITE4(sc, YMF_MODE, YMF_MODE_RESET);
        YWRITE4(sc, YMF_MODE, 0);

        YWRITE4(sc, YMF_MAPOF_REC, 0);
        YWRITE4(sc, YMF_MAPOF_EFFECT, 0);
        YWRITE4(sc, YMF_PLAY_CTRLBASE, 0);
        YWRITE4(sc, YMF_REC_CTRLBASE, 0);
        YWRITE4(sc, YMF_EFFECT_CTRLBASE, 0);
        YWRITE4(sc, YMF_WORK_BASE, 0);

        ctrl = YREAD2(sc, YMF_GLOBAL_CONTROL);
        YWRITE2(sc, YMF_GLOBAL_CONTROL,
		ctrl & ~0x0007);

	/* download DSP microcode. */
	p = ymf_dsp_mcode;
	size = sizeof(ymf_dsp_mcode);
	YWRITEREGION4(sc, YMF_DSP_INSTRAM, p, size);

	/* download CONTROL microcode. */
	p = ctrls[dstype].mcode;
	size = ctrls[dstype].size;
	YWRITEREGION4(sc, YMF_CTRL_INSTRAM, p, size);

	ymf_enable_dsp(sc);
	delay(10*1000);		/* XXX */

	return 0;
}

static int
ymf_allocate_slots(sc)
	struct ymf_softc *sc;
{
	size_t pcs, rcs, ecs, ws, memsize;
	void *mp;
	off_t cb;
	int i;
	char *t;
	bus_dma_segment_t segs;
	int rsegs;

	/* alloc DSP Control Data */
	pcs = YREAD4(sc, YMF_PLAY_CTRLSIZE) * sizeof(u_int32_t);
	rcs = YREAD4(sc, YMF_REC_CTRLSIZE) * sizeof(u_int32_t);
	ecs = YREAD4(sc, YMF_EFFECT_CTRLSIZE) * sizeof(u_int32_t);
	ws  = YREAD4(sc, YMF_WORK_SIZE) * sizeof(u_int32_t);

	DPRINTF(("play control size : %d\n",pcs));
	DPRINTF(("rec control size : %d\n",rcs));
	DPRINTF(("eff control size : %d\n",ecs));
	DPRINTF(("work size : %d\n",ws));

	memsize = 64*2*pcs + 2*2*rcs + 5*2*ecs + ws;
	memsize += (64+1)*4;	/* play control data table */

	i = bus_dmamem_alloc(sc->sc_dmatag, memsize, 0, 0, &segs, 1, &rsegs,
			     BUS_DMA_NOWAIT);
	if (i != 0) {
		printf("%s: couldn't alloc DSP Control Data, reason %d\n",
		       sc->sc_dev.dv_xname, i);
		return 1;
	}
	bus_dmamem_map(sc->sc_dmatag,
		       &segs, rsegs, memsize, (caddr_t*) &mp,
		       BUS_DMA_NOWAIT|BUS_DMA_COHERENT);

	DPRINTF(("mp:%p, vtophys(mp):%p\n",mp,(void*)vtophys((vaddr_t)mp)));

	bzero(mp, memsize);
	sc->ctrlbase = vtophys((vaddr_t)mp);

        cb = 0;
	t = (u_int8_t*)mp;
	YWRITE4(sc, YMF_WORK_BASE, sc->ctrlbase + cb);

        cb += ws;
        sc->pbase = (u_int32_t *)(t + cb);
        /* printf("pbase = %p -> 0x%x\n", sc->pbase, sc->ctrlbase + cb); */
        YWRITE4(sc, YMF_PLAY_CTRLBASE, sc->ctrlbase + cb);
        cb += (64 + 1) * 4;
        sc->rbank = (struct rec_slot_ctrl_bank *)(t + cb);
        YWRITE4(sc, YMF_REC_CTRLBASE, sc->ctrlbase + cb);
        cb += 2 * 2 * rcs;
        YWRITE4(sc, YMF_EFFECT_CTRLBASE, sc->ctrlbase + cb);
        cb += 5 * 2 * ecs;

        sc->pbankbase = sc->ctrlbase + cb;
        for (i=0; i<64; i++) {
		sc->pbank[i*2] = (struct play_slot_ctrl_bank *)(t + cb);
		*(sc->pbase + i+1) = vtophys((vaddr_t)sc->pbank[i*2]);
                cb += pcs;

                sc->pbank[i*2+1] = (struct play_slot_ctrl_bank *)(t + cb);
                cb += pcs;
        }

	return 0;
}

static void
ymf_enable_dsp(sc)
	struct ymf_softc *sc;
{
	YWRITE4(sc, YMF_CONFIG, YMF_DSP_SETUP);
}

static int
ymf_disable_dsp(sc)
	struct ymf_softc *sc;
{
	int to;
	u_int32_t data;

	data = YREAD4(sc, YMF_CONFIG);
	if (data)
		YWRITE4(sc, YMF_CONFIG, YMF_DSP_DISABLE);

	for (to = 0; to < YMF_WORK_TIMEOUT; to++) {
		if ((YREAD4(sc, YMF_STATUS) & YMF_STAT_WORK) == 0)
			return 0;
		delay(1);
	}

	return 1;
}

int
ymf_match(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct pci_attach_args *pa = (struct pci_attach_args *) aux;

	switch (PCI_VENDOR(pa->pa_id)) {
	case PCI_VENDOR_YAMAHA:
		switch (PCI_PRODUCT(pa->pa_id)) {
		case PCI_PRODUCT_YAMAHA_YMF724:
		case PCI_PRODUCT_YAMAHA_YMF740:
		case PCI_PRODUCT_YAMAHA_YMF740C:
		case PCI_PRODUCT_YAMAHA_YMF724F:
		case PCI_PRODUCT_YAMAHA_YMF744B:
		case PCI_PRODUCT_YAMAHA_YMF754:
			return (1);
		}
		break;
	}

	return (0);
}

/*
 * This routine is called after all the ISA devices are configured,
 * to avoid conflict.
 */
static void
ymf_configure_legacy (arg)
	struct device *arg;
{
	struct ymf_softc *sc = (struct ymf_softc*) arg;
	pcireg_t reg;
	struct device *dev;
	int type;
	int i;
	bus_addr_t opl_addrs[] = {0x388, 0x398, 0x3A0, 0x3A8};
	bus_addr_t mpu_addrs[] = {0x330, 0x300, 0x332, 0x334};

	switch (PCI_PRODUCT(sc->sc_id)) {
		case PCI_PRODUCT_YAMAHA_YMF724F:
			type = 0;
			break;
#if 1				/* XXX not tested */
		case PCI_PRODUCT_YAMAHA_YMF754:
			type = 1;
			break;
#endif
		/* I do not know other case */
		default:
			return;
	}

	reg = pci_conf_read(sc->sc_pc, sc->sc_pcitag, YMF_PCI_LEGACY);
	reg &= ~0x8133c03f;	/* these bits are out of interest */
	reg |= ((YMF_PCI_EX_LEGACY_IMOD) |
		(YMF_PCI_LEGACY_FMEN |
		 YMF_PCI_LEGACY_MEN /*| YMF_PCI_LEGACY_MIEN*/));
	if (type == 1) {
		pci_conf_write(sc->sc_pc, sc->sc_pcitag, YMF_PCI_LEGACY, reg);
		delay(100*1000);
	}

	/* Try for OPL */
	dev = 0;
	for (i = 0; i < sizeof(opl_addrs) / sizeof (bus_addr_t); i++) {
		if (type == 0) {
			pci_conf_write(sc->sc_pc, sc->sc_pcitag,
				       YMF_PCI_LEGACY, reg | (i << (0+16)));
			delay(100*1000);	/* wait 100ms */
		} else
			pci_conf_write(sc->sc_pc, sc->sc_pcitag,
				       YMF_PCI_FM_BA, opl_addrs[i]);
		if (bus_space_map(sc->sc_opl_iot,
				  opl_addrs[i], 4, 0, &sc->sc_opl_ioh) == 0) {
			struct audio_attach_args aa; 

			aa.type = AUDIODEV_TYPE_OPL;
			aa.hwif = aa.hdl = NULL;
			dev = config_found(&sc->sc_dev, &aa, audioprint);
			if (dev == 0)
				bus_space_unmap(sc->sc_opl_iot,
						sc->sc_opl_ioh, 4);
			else {
				if (type == 0)
					reg |= (i << (0+16));
				break;
			}
		} 
	}
	if (dev == 0) {
		reg &= ~YMF_PCI_LEGACY_FMEN;
		pci_conf_write(sc->sc_pc, sc->sc_pcitag,
			       YMF_PCI_LEGACY, reg);
	}

	/* Try for MPU */
	dev = 0;
	for (i = 0; i < sizeof(mpu_addrs) / sizeof (bus_addr_t); i++) {
		if (type == 0)
			pci_conf_write(sc->sc_pc, sc->sc_pcitag,
				       YMF_PCI_LEGACY, reg | (i << (4+16)));
		else
			pci_conf_write(sc->sc_pc, sc->sc_pcitag,
				       YMF_PCI_MPU_BA, mpu_addrs[i]);
		if (bus_space_map(sc->sc_mpu_iot,
				  mpu_addrs[i], 2, 0, &sc->sc_mpu_ioh) == 0) {
			struct audio_attach_args aa; 

			aa.type = AUDIODEV_TYPE_MPU;
			aa.hwif = aa.hdl = NULL;
			dev = config_found(&sc->sc_dev, &aa, audioprint);
			if (dev == 0)
				bus_space_unmap(sc->sc_mpu_iot,
						sc->sc_mpu_ioh, 2);
			else {
				if (type == 0)
					reg |= (i << (4+16));
				break;
			}
		}
	}
	if (dev == 0) {
		reg &= ~(YMF_PCI_LEGACY_MEN | YMF_PCI_LEGACY_MIEN);
		pci_conf_write(sc->sc_pc, sc->sc_pcitag,
			       YMF_PCI_LEGACY, reg);
	}
	sc->sc_mpu = dev;
} 

void
ymf_attach(parent, self, aux)
	struct device *parent;
	struct device *self;
	void *aux;
{
	struct ymf_softc *sc = (struct ymf_softc *)self;
	struct pci_attach_args *pa = (struct pci_attach_args *)aux;
	pci_chipset_tag_t pc = pa->pa_pc;
	char const *intrstr;
	pci_intr_handle_t ih;
	pcireg_t reg;
	struct ymf_codec_softc *codec;
	char devinfo[256];
	mixer_ctrl_t ctl;
	int i, r, to;
	int revision;
	int ac97_id2;

	pci_devinfo(pa->pa_id, pa->pa_class, 0, devinfo);
	revision = PCI_REVISION(pa->pa_class);
	printf(": %s (rev. 0x%02x)\n", devinfo, revision);

	/* Map register to memory */
	if (pci_mapreg_map(pa, YMF_PCI_MBA, PCI_MAPREG_TYPE_MEM, 0,
	    &sc->memt, &sc->memh, NULL, NULL)) {
		printf("%s: can't map memory space\n", sc->sc_dev.dv_xname);
		return;
	}

	/* Map and establish the interrupt. */
	if (pci_intr_map(pa, &ih)) {
		printf("%s: couldn't map interrupt\n", sc->sc_dev.dv_xname);
		return;
	}
	intrstr = pci_intr_string(pc, ih);
	sc->sc_ih = pci_intr_establish(pc, ih, IPL_AUDIO, ymf_intr, sc);
	if (sc->sc_ih == NULL) {
		printf("%s: couldn't establish interrupt",
		    sc->sc_dev.dv_xname);
		if (intrstr != NULL)
			printf(" at %s", intrstr);
		printf("\n");
		return;
	}
	printf("%s: interrupting at %s\n", sc->sc_dev.dv_xname, intrstr);

	sc->sc_dmatag = pa->pa_dmat;
	sc->sc_pc = pc;
	sc->sc_pcitag = pa->pa_tag;
	sc->sc_id = pa->pa_id;

	/* disable legacy mode */
	reg = pci_conf_read(pc, pa->pa_tag, YMF_PCI_LEGACY);
	pci_conf_write(pc, pa->pa_tag, YMF_PCI_LEGACY,
		       reg & YMF_PCI_LEGACY_LAD);

	/* Enable the device. */
	reg = pci_conf_read(pc, pa->pa_tag, PCI_COMMAND_STATUS_REG);
	reg |= (PCI_COMMAND_IO_ENABLE | PCI_COMMAND_MEM_ENABLE |
		PCI_COMMAND_MASTER_ENABLE);
	pci_conf_write(pc, pa->pa_tag, PCI_COMMAND_STATUS_REG, reg);
	reg = pci_conf_read(pc, pa->pa_tag, PCI_COMMAND_STATUS_REG);

	/* download microcode. */
	if (ymf_download_mcode(sc)) {
		printf("%s: download microcode failed\n", sc->sc_dev.dv_xname);
		return;
	}
	if (ymf_allocate_slots(sc)) {
		printf("%s: could not allocate slots\n", sc->sc_dev.dv_xname);
		return;
	}

	reg = pci_conf_read(pc, pa->pa_tag, YMF_PCI_DSCTRL);
	pci_conf_write(pc, pa->pa_tag, YMF_PCI_DSCTRL, reg | YMF_DSCTRL_WRST);
	delay(50000);

	/*
	 * detect Primary/Secondary AC97
	 *	YMF754 Hardware Specification Rev 1.01 page 24
	 */
	reg = pci_conf_read(pc, pa->pa_tag, YMF_PCI_DSCTRL);
	pci_conf_write(pc, pa->pa_tag, YMF_PCI_DSCTRL,
		reg & ~YMF_DSCTRL_CRST);

	/* Primary */
	for (to = 0; to < AC97_TIMEOUT; to++) {
		if ((YREAD2(sc, AC97_STAT_ADDR1) & AC97_BUSY) == 0)
			break;
		delay(1);
	}
	if (to == AC97_TIMEOUT) {
		printf("%s: no AC97 avaliable\n", sc->sc_dev.dv_xname);
		return;
	}

	/* Secondary */
	ac97_id2 = -1;
	if ((YREAD2(sc, YMF_ACTIVITY) & YMF_ACTIVITY_DOCKA) == 0)
		goto detected;
#if 0
	YWRITE2(sc, YMF_GPIO_OCTRL,
		YREAD2(sc, YMF_GPIO_OCTRL) & ~YMF_GPIO_GPO2);
	YWRITE2(sc, YMF_GPIO_FUNCE,
		(YREAD2(sc, YMF_GPIO_FUNCE)&(~YMF_GPIO_GPC2))|YMF_GPIO_GPE2);
#endif
	for (to = 0; to < AC97_TIMEOUT; to++) {
		if ((YREAD2(sc, AC97_STAT_ADDR2) & AC97_BUSY) == 0)
			break;
		delay(1);
	}
	if (to < AC97_TIMEOUT) {
		/* detect id */
		for (ac97_id2 = 1; ac97_id2 < 4; ac97_id2++) {
			YWRITE2(sc, AC97_CMD_ADDR,
				AC97_CMD_READ | AC97_ID(ac97_id2) | 0x28);

			for (to = 0; to < AC97_TIMEOUT; to++) {
				if ((YREAD2(sc, AC97_STAT_ADDR2) & AC97_BUSY)
				    == 0)
					goto detected;
				delay(1);
			}
		}
		if (ac97_id2 == 4)
			ac97_id2 = -1;
detected:
	}

	pci_conf_write(pc, pa->pa_tag, YMF_PCI_DSCTRL,
		       reg | YMF_DSCTRL_CRST);
	delay (20);
	pci_conf_write(pc, pa->pa_tag, YMF_PCI_DSCTRL,
		reg & ~YMF_DSCTRL_CRST);
	for (to = 0; to < AC97_TIMEOUT; to++) {
		if ((YREAD2(sc, AC97_STAT_ADDR1) & AC97_BUSY) == 0)
			break;
		delay(1);
	}

	/*
	 * attach ac97 codec
	 */
	for (i = 0; i < 2; i++) {
		static struct {
			int data;
			int addr;
		} statregs[] = {
			{AC97_STAT_DATA1, AC97_STAT_ADDR1},
			{AC97_STAT_DATA2, AC97_STAT_ADDR2},
		};

		if (i == 1 && ac97_id2 == -1)
			break;		/* secondary ac97 not available */

		codec = &sc->sc_codec[i];
		memcpy(&codec->sc_dev, &sc->sc_dev, sizeof(codec->sc_dev));
		codec->sc = sc;
		codec->id = i == 1 ? ac97_id2 : 0;
		codec->status_data = statregs[i].data;
		codec->status_addr = statregs[i].addr;
		codec->host_if.arg = codec;
		codec->host_if.attach = ymf_attach_codec;
		codec->host_if.read = ymf_read_codec;
		codec->host_if.write = ymf_write_codec;
		codec->host_if.reset = ymf_reset_codec;

		if ((r = ac97_attach(&codec->host_if)) != 0) {
			printf("%s: can't attach codec (error 0x%X)\n",
				sc->sc_dev.dv_xname, r);
			return;
		}
	}

	/* Just enable the DAC and master volumes by default */
	ctl.type = AUDIO_MIXER_ENUM;
	ctl.un.ord = 0;  /* off */
	ctl.dev = ymf_get_portnum_by_name(sc, AudioCoutputs,
	       AudioNmaster, AudioNmute);
	ymf_mixer_set_port(sc, &ctl);
	ctl.dev = ymf_get_portnum_by_name(sc, AudioCinputs,
	       AudioNdac, AudioNmute);
	ymf_mixer_set_port(sc, &ctl);
	ctl.dev = ymf_get_portnum_by_name(sc, AudioCrecord,
	       AudioNvolume, AudioNmute);
	ymf_mixer_set_port(sc, &ctl);
	
	ctl.dev = ymf_get_portnum_by_name(sc, AudioCrecord,
	       AudioNsource, NULL);
	ctl.type = AUDIO_MIXER_ENUM;
	ctl.un.ord = 0;
	ymf_mixer_set_port(sc, &ctl);

	/* set a reasonable default volume */
	ctl.type = AUDIO_MIXER_VALUE;
	ctl.un.value.num_channels = 2;
	ctl.un.value.level[AUDIO_MIXER_LEVEL_LEFT] =
	ctl.un.value.level[AUDIO_MIXER_LEVEL_RIGHT] = 127;

	ctl.dev = sc->sc_codec[0].codec_if->vtbl->get_portnum_by_name(
		sc->sc_codec[0].codec_if, AudioCoutputs, AudioNmaster, NULL);
	ymf_mixer_set_port(sc, &ctl);

	for (i = 0x80; i < 0xc0; i += 2)
		YWRITE2(sc, i, 0x3fff);

	audio_attach_mi(&ymf_hw_if, sc, &sc->sc_dev);

	sc->sc_legacy_iot = pa->pa_iot;
	config_defer((struct device*) sc, ymf_configure_legacy);
}

int
ymf_attach_codec(sc_, codec_if)
	void *sc_;
	struct ac97_codec_if *codec_if;
{
	struct ymf_codec_softc *sc = sc_;

	sc->codec_if = codec_if;
	return 0;
}

static int
ymf_ready_codec(sc)
	struct ymf_codec_softc *sc;
{
	int to;

	for (to = 0; to < AC97_TIMEOUT; to++) {
		if ((YREAD2(sc->sc, sc->status_addr) & AC97_BUSY) == 0)
			return 0;
		delay(1);
	}

	return 1;
}

int
ymf_read_codec(sc_, reg, data)
	void *sc_;
	u_int8_t reg;
	u_int16_t *data;
{
	struct ymf_codec_softc *sc = sc_;

	YWRITE2(sc->sc, AC97_CMD_ADDR, AC97_CMD_READ | AC97_ID(sc->id) | reg);

	if (ymf_ready_codec(sc)) {
		printf("%s: ymf_read_codec timeout\n",
		       sc->sc->sc_dev.dv_xname);
		return EIO;
	}

	*data = YREAD2(sc->sc, sc->status_data);

	return 0;
}

int
ymf_write_codec(sc_, reg, data)
	void *sc_;
	u_int8_t reg;
	u_int16_t data;
{
	struct ymf_codec_softc *sc = sc_;

	YWRITE2(sc->sc, AC97_CMD_ADDR, AC97_CMD_WRITE | AC97_ID(sc->id) | reg);
	YWRITE2(sc->sc, AC97_CMD_DATA, data);

	if (ymf_ready_codec(sc)) {
		printf("%s: ymf_write_codec timeout\n",
			sc->sc->sc_dev.dv_xname);
		return EIO;
	}

	return 0;
}

/*
 * XXX: Must handle secondary differntly!!
 */
void
ymf_reset_codec(sc_)
	void *sc_;
{
	struct ymf_softc *sc = sc_;
	pcireg_t reg;

	/* reset AC97 codec */
	reg = pci_conf_read(sc->sc_pc, sc->sc_pcitag, YMF_PCI_DSCTRL);
	if (reg & 0x03) {
		pci_conf_write(sc->sc_pc, sc->sc_pcitag,
			       YMF_PCI_DSCTRL, reg & ~0x03);
		pci_conf_write(sc->sc_pc, sc->sc_pcitag,
			       YMF_PCI_DSCTRL, reg | 0x03);
		pci_conf_write(sc->sc_pc, sc->sc_pcitag,
			       YMF_PCI_DSCTRL, reg & ~0x03);
		delay(50000);
	}

	ymf_ready_codec(sc_);
}

int
ymf_intr(p)
	void *p;
{
	struct ymf_softc *sc = p;
	u_int status;
	int bank = (sc->sc_bank == 0);

	status = YREAD4(sc, YMF_STATUS);
	if ((status & (YMF_STAT_INT|YMF_STAT_TINT)) == 0) {
#if NMPU > 0
		if (sc->sc_mpu)
			return mpu_intr(sc->sc_mpu);
#endif
		return 0;
	}

	/* INT/TINT clear */
	YWRITE4(sc, YMF_STATUS, YMF_STAT_INT | YMF_STAT_TINT);

#if 0
	if (status & YMF_STAT_TINT) {
		printf ("ymf_intr: timeout\n");
		/* timeout */
	}
#endif
	if ((status & YMF_STAT_INT) == 0) {
		return 1;
	}

	if (sc->sc_play.intr) {
		u_int dma = sc->pbank[bank]->pgstart * sc->sc_play.factor;
		u_int cpu = sc->sc_play.offset;
		u_int blksize = sc->sc_play.blksize;
		u_int mode;

#if 0
		sc->pbank[bank]->pgbase =
		sc->pbank[bank+2]->pgbase = sc->sc_play.base;
		sc->pbank[sc->sc_bank]->pgbase =
		sc->pbank[sc->sc_bank+2]->pgbase = 0;
#endif
		mode = YREAD4(sc, YMF_MODE);
		YWRITE4(sc, YMF_MODE, mode | YMF_MODE_ACTV2);

		if (((dma > cpu) && (dma - cpu > blksize)) ||
		    ((cpu > dma) &&
		     (dma + sc->sc_play.length - cpu > blksize))) {
			/* we can fill the next block */
			sc->sc_play.offset = dma;
			sc->sc_play.intr(sc->sc_play.intr_arg);
		}
	}
	if (sc->sc_rec.intr) {
		u_int dma = sc->rbank[bank].pgstartadr * sc->sc_rec.factor;
		u_int cpu = sc->sc_rec.offset;
		u_int blksize = sc->sc_rec.blksize;
		u_int mode;

		sc->rbank[bank].pgbase =
		sc->rbank[bank+2].pgbase = sc->sc_rec.base;
		sc->rbank[sc->sc_bank].pgbase =
		sc->rbank[sc->sc_bank+2].pgbase = 0;
		mode = YREAD4(sc, YMF_MODE);
		YWRITE4(sc, YMF_MODE, mode | YMF_MODE_ACTV2);

		if (((dma > cpu) && (dma - cpu >= blksize)) ||
		    ((cpu > dma) &&
		     (dma + sc->sc_rec.length - cpu > blksize))) {
			/* we can fill the next block */
			sc->sc_rec.offset = dma;
			sc->sc_rec.intr(sc->sc_rec.intr_arg);
		}
	}

	YWRITE4(sc, YMF_CONTROL_SELECT, bank);
	sc->sc_bank = bank; 

	return 1;
}

int
ymf_allocmem(sc, size, align, p)
	struct ymf_softc *sc;
	size_t size;
	size_t align;
	struct ymf_dma *p;
{
	int error;

	p->size = size;
	error = bus_dmamem_alloc(sc->sc_dmatag, p->size, align, 0,
				 p->segs, sizeof(p->segs)/sizeof(p->segs[0]),
				 &p->nsegs, BUS_DMA_NOWAIT);
	if (error)
		return (error);

	error = bus_dmamem_map(sc->sc_dmatag, p->segs, p->nsegs, p->size, 
			       &p->addr, BUS_DMA_NOWAIT|BUS_DMA_COHERENT);
	if (error)
		goto free;

	error = bus_dmamap_create(sc->sc_dmatag, p->size, 1, p->size,
				  0, BUS_DMA_NOWAIT, &p->map);
	if (error)
		goto unmap;

	error = bus_dmamap_load(sc->sc_dmatag, p->map, p->addr, p->size, NULL, 
				BUS_DMA_NOWAIT);
	if (error)
		goto destroy;
	return (0);

destroy:
	bus_dmamap_destroy(sc->sc_dmatag, p->map);
unmap:
	bus_dmamem_unmap(sc->sc_dmatag, p->addr, p->size);
free:
	bus_dmamem_free(sc->sc_dmatag, p->segs, p->nsegs);
	return (error);
}

int
ymf_freemem(sc, p)
	struct ymf_softc *sc;
	struct ymf_dma *p;
{
	bus_dmamap_unload(sc->sc_dmatag, p->map);
	bus_dmamap_destroy(sc->sc_dmatag, p->map);
	bus_dmamem_unmap(sc->sc_dmatag, p->addr, p->size);
	bus_dmamem_free(sc->sc_dmatag, p->segs, p->nsegs);
	return 0;
}

int
ymf_open(addr, flags)
	void *addr;
	int flags;
{
	struct ymf_softc *sc = addr;
	int mode;

	mode = YREAD4(sc, YMF_MODE);
	mode |= YMF_MODE_ACTV;
	mode &= ~YMF_MODE_ACTV2;
	YWRITE4(sc, YMF_MODE, mode);

	return 0;
}

/*
 * Close function is called at splaudio().
 */
void
ymf_close(addr)
	void *addr;
{
	struct ymf_softc *sc = addr;

	ymf_halt_output(sc);
	ymf_halt_input(sc);
	ymf_halt(sc);
}

int
ymf_query_encoding(addr, fp)
	void *addr;
	struct audio_encoding *fp;
{
	switch (fp->index) {
	case 0:
		strcpy(fp->name, AudioEulinear);
		fp->encoding = AUDIO_ENCODING_ULINEAR;
		fp->precision = 8;
		fp->flags = 0;
		return (0);
	case 1:
		strcpy(fp->name, AudioEmulaw);
		fp->encoding = AUDIO_ENCODING_ULAW;
		fp->precision = 8;
		fp->flags = AUDIO_ENCODINGFLAG_EMULATED;
		return (0);
	case 2:
		strcpy(fp->name, AudioEalaw);
		fp->encoding = AUDIO_ENCODING_ALAW;
		fp->precision = 8;
		fp->flags = AUDIO_ENCODINGFLAG_EMULATED;
		return (0);
	case 3:
		strcpy(fp->name, AudioEslinear);
		fp->encoding = AUDIO_ENCODING_SLINEAR;
		fp->precision = 8;
		fp->flags = AUDIO_ENCODINGFLAG_EMULATED;
		return (0);
	case 4:
		strcpy(fp->name, AudioEslinear_le);
		fp->encoding = AUDIO_ENCODING_SLINEAR_LE;
		fp->precision = 16;
		fp->flags = 0;
		return (0);
	case 5:
		strcpy(fp->name, AudioEulinear_le);
		fp->encoding = AUDIO_ENCODING_ULINEAR_LE;
		fp->precision = 16;
		fp->flags = AUDIO_ENCODINGFLAG_EMULATED;
		return (0);
	case 6:
		strcpy(fp->name, AudioEslinear_be);
		fp->encoding = AUDIO_ENCODING_SLINEAR_BE;
		fp->precision = 16;
		fp->flags = AUDIO_ENCODINGFLAG_EMULATED;
		return (0);
	case 7:
		strcpy(fp->name, AudioEulinear_be);
		fp->encoding = AUDIO_ENCODING_ULINEAR_BE;
		fp->precision = 16;
		fp->flags = AUDIO_ENCODINGFLAG_EMULATED;
		return (0);
	default:
		return (EINVAL);
	}
}

int
ymf_set_params(addr, setmode, usemode, play, rec)
	void *addr;
	int setmode, usemode;
	struct audio_params *play, *rec;
{
	struct audio_params *p;
	int mode;

	for (mode = AUMODE_RECORD; mode != -1; 
	     mode = mode == AUMODE_RECORD ? AUMODE_PLAY : -1) {
		if ((setmode & mode) == 0)
			continue;

		p = mode == AUMODE_PLAY ? play : rec;

		if (p->sample_rate < 4000 || p->sample_rate > 48000 ||
		    (p->precision != 8 && p->precision != 16) ||
		    (p->channels != 1 && p->channels != 2))
			return (EINVAL);

		p->factor = 1;
		p->sw_code = 0;
		switch (p->encoding) {
		case AUDIO_ENCODING_SLINEAR_BE:
			if (p->precision == 16)
				p->sw_code = swap_bytes;
			else
				p->sw_code = change_sign8;
			break;
		case AUDIO_ENCODING_SLINEAR_LE:
			if (p->precision != 16)
				p->sw_code = change_sign8;
			break;
		case AUDIO_ENCODING_ULINEAR_BE:
			if (p->precision == 16) {
				if (mode == AUMODE_PLAY)
					p->sw_code = swap_bytes_change_sign16_le;
				else
					p->sw_code = change_sign16_swap_bytes_le;
			}
			break;
		case AUDIO_ENCODING_ULINEAR_LE:
			if (p->precision == 16)
				p->sw_code = change_sign16_le;
			break;
		case AUDIO_ENCODING_ULAW:
			if (mode == AUMODE_PLAY) {
				p->factor = 2;
				p->precision = 16;
				p->sw_code = mulaw_to_slinear16_le;
			} else
				p->sw_code = ulinear8_to_mulaw;
			break;
		case AUDIO_ENCODING_ALAW:
			if (mode == AUMODE_PLAY) {
				p->factor = 2;
				p->precision = 16;
				p->sw_code = alaw_to_slinear16_le;
			} else
				p->sw_code = ulinear8_to_alaw;
			break;
		default:
			return (EINVAL);
		}
	}
	return 0;
}

int
ymf_round_blocksize(addr, blk)
	void *addr;
	int blk;
{
	/*
	 * Block size must be bigger than a frame.
	 * That is 1024bytes at most, i.e. for 48000Hz, 16bit, 2ch.
	 */
	if (blk < 1024)
		blk = 1024;
	return (blk & -4);	/* keep good alignment */
}

static u_int32_t
ymf_get_lpfq(sample_rate)
	u_int sample_rate;
{
	int i;
	static struct lpfqt {
		u_int rate;
		u_int32_t lpfq;
	} lpfqt[] = {
		{8000,  0x32020000},
		{11025, 0x31780000},
		{16000, 0x31380000},
		{22050, 0x31c80000},
		{32000, 0x33d00000},
		{44100, 0x40000000},
		{48000, 0x40000000},
		{0, 0}
	};

	for (i = 0; lpfqt[i].rate != 0; i++)
		if (sample_rate <= lpfqt[i].rate)
			break;

	return lpfqt[i].lpfq;
}

static u_int32_t
ymf_get_lpfk(sample_rate)
	u_int sample_rate;
{
	int i;
	static struct lpfkt {
		u_int rate;
		u_int32_t lpfk;
	} lpfkt[] = {
		{8000,  0x18b00000},
		{11025, 0x20900000},
		{16000, 0x2b980000},
		{22050, 0x35a00000},
		{32000, 0x40000000},
		{44100, 0x40000000},
		{48000, 0x40000000},
		{0, 0}
	};

	for (i = 0; lpfkt[i].rate != 0; i++)
		if (sample_rate <= lpfkt[i].rate)
			break;

	return lpfkt[i].lpfk;
}

int
ymf_trigger_output(addr, start, end, blksize, intr, arg, param)
	void *addr;
	void *start, *end;
	int blksize;
	void (*intr) __P((void *));
	void *arg;
	struct audio_params *param;
{
	struct ymf_softc *sc = addr;
	struct ymf_dma *p;
	struct play_slot_ctrl_bank *psb;
	const u_int gain = 0x40000000;
	bus_addr_t s;
	size_t l;
	int i, bank;

#ifdef DIAGNOSTIC
	if (sc->sc_play.intr)
		panic("ymf_trigger_output: already running");
#endif

	sc->sc_play.intr = intr;
	sc->sc_play.intr_arg = arg;
	sc->sc_play.offset = 0;
	sc->sc_play.blksize = blksize;

	YWRITE4(sc, YMF_CONTROL_SELECT, sc->sc_bank);

	DPRINTFN(1, ("ymf_trigger_output: sc=%p start=%p end=%p "
	    "blksize=%d intr=%p(%p)\n", addr, start, end, blksize, intr, arg));

	p = ymf_find_dma(sc, start);
	if (!p) {
		printf("ymf_trigger_output: bad addr %p\n", start);
		return (EINVAL);
	}

#ifdef DIAGNOSTIC
	{
		u_int32_t ctrlsize;
		if ((ctrlsize = YREAD4(sc, YMF_PLAY_CTRLSIZE)) !=
		    sizeof(struct play_slot_ctrl_bank) / sizeof(u_int32_t))
			panic("%s: invalid play slot ctrldata %d %d",
				sc->sc_dev.dv_xname, ctrlsize, sizeof(struct play_slot));
	}
#endif

	s = DMAADDR(p);
	l = ((char *)end - (char *)start);
	sc->sc_play.base = s;
	sc->sc_play.length = l;

	*sc->pbase = param->channels;		/* Num of play */

	sc->sc_play.factor = 1;
	if (param->channels == 2)
		sc->sc_play.factor *= 2;
	if (param->precision != 8)
		sc->sc_play.factor *= 2;
	l /= sc->sc_play.factor;

	psb = sc->pbank[0];
	psb->format = ((param->channels == 2 ? PSLT_FORMAT_STEREO : 0) |
		      (param->precision == 8 ? PSLT_FORMAT_8BIT : 0) |
		      (param->sample_rate == 44100 ?
		       PSLT_FORMAT_SRC441 : 0));
	psb->loopdefault = 0;
	psb->pgbase = 0;
	psb->pgloop = 0;
	psb->pgloopend = l+1;
	psb->pgloopfrac = 0;
	psb->pgdeltaend = (param->sample_rate * 65536 / 48000) << 12;
	psb->lpfkend = ymf_get_lpfk(param->sample_rate);
	psb->eggainend = psb->eggainend = gain;
	psb->lchgainend = psb->rchgainend = 0;
	psb->effect1gainend = psb->effect2gainend = psb->effect3gainend = 0;
	psb->lpfq = ymf_get_lpfq(param->sample_rate);
	psb->status = 0;
	psb->numofframes = 0;
	psb->loopcount = 0;
	psb->pgstart = psb->pgstartfrac = 0;
	psb->pgdelta = psb->pgdeltaend;
	psb->lpfk = ymf_get_lpfk(param->sample_rate);
	psb->eggain = gain;
	psb->lchgain = psb->rchgain = 0;
	psb->effect1gain = psb->effect2gain = psb->effect3gain = 0;
	psb->lpfd1 = psb->lpfd2 = 0;

	for (i = 0; i < param->channels; i++) {
		for (bank = 0; bank < N_PLAY_SLOT_BANK; bank++) {
			psb = sc->pbank[i*2+bank];
			if (i || bank)
				*psb = *(sc->pbank[0]);
			if (param->channels == 2) {
				if (i == 0) {
					psb->lchgain = psb->lchgainend = gain;
				} else {
					psb->rchgain = psb->rchgainend = gain;
					psb->format |= PSLT_FORMAT_RCH;
				}
			} else {
				psb->lchgain = psb->lchgainend = gain;
				psb->rchgain = psb->rchgainend = gain;
			}
#if 0
			if (bank == sc->sc_bank)
#endif
				psb->pgbase = s;
		}
	}

	YMF_DUMP_PLAY_SLOT(5, sc, 0);
	YMF_DUMP_PLAY_SLOT(5, sc, 1);

	YWRITE4(sc, YMF_CONTROL_SELECT, 0);
	YWRITE4(sc, YMF_MODE,
		YREAD4(sc, YMF_MODE) | YMF_MODE_ACTV | YMF_MODE_ACTV2);

	return 0;
}

int
ymf_trigger_input(addr, start, end, blksize, intr, arg, param)
	void *addr;
	void *start, *end;
	int blksize;
	void (*intr) __P((void *));
	void *arg;
	struct audio_params *param;
{
	struct ymf_softc *sc = addr;
	struct ymf_dma *p;
	u_int data;
	struct rec_slot_ctrl_bank *rsb;
	bus_addr_t s;
	size_t l;

#ifdef DIAGNOSTIC
	if (sc->sc_rec.intr)
		panic("ymf_trigger_input: already running");
#endif
	sc->sc_rec.intr = intr;
	sc->sc_rec.intr_arg = arg;
	sc->sc_rec.offset = 0;
	sc->sc_rec.blksize = blksize;

	DPRINTFN(1, ("ymf_trigger_input: "
	    "sc=%p start=%p end=%p blksize=%d intr=%p(%p)\n", 
	    addr, start, end, blksize, intr, arg));

	p = ymf_find_dma(sc, start);
	if (!p) {
		printf("ymf_trigger_input: bad addr %p\n", start);
		return (EINVAL);
	}

#ifdef DIAGNOSTIC
	{
		u_int32_t ctrlsize;
		if ((ctrlsize = YREAD4(sc, YMF_REC_CTRLSIZE)) !=
		    sizeof(struct rec_slot) / sizeof(u_int32_t))
			panic("%s: invalid rec slot ctrldata %d",
				sc->sc_dev.dv_xname, ctrlsize);
	}
#endif

	s = DMAADDR(p);
	l = ((char *)end - (char *)start);
	sc->sc_rec.base = s;
	sc->sc_rec.length = l;

	sc->sc_rec.factor = 1;
	if (param->channels == 2)
		sc->sc_rec.factor *= 2;
	if (param->precision != 8)
		sc->sc_rec.factor *= 2;
	l /= sc->sc_rec.factor;

	rsb = &sc->rbank[0];
	rsb->pgbase = s;
	rsb->pgloopendadr = l;
	rsb->pgstartadr = 0;
	rsb->numofloops = 0;

	YWRITE4(sc, YMF_CONTROL_SELECT, sc->sc_bank);
	YWRITE4(sc, YMF_MAPOF_REC, YMF_ADCSLOT_VALID | YMF_RECSLOT_VALID);

	data = 48000 * 4096 / param->sample_rate - 1;
	YWRITE4(sc, YMF_ADC_SAMPLE_RATE, data);
	YWRITE4(sc, YMF_REC_SAMPLE_RATE, data);

	data = (param->precision == 8 ? YMF_FORMAT_8BIT : 0) |
	       (param->channels == 2 ? YMF_FORMAT_STEREO : 0);
	YWRITE4(sc, YMF_ADC_FORMAT, data);
	YWRITE4(sc, YMF_REC_FORMAT, data);

	return 0;
}

static int
ymf_halt(sc)
	struct ymf_softc *sc;
{
	u_int32_t mode;

	mode = YREAD4(sc, YMF_MODE);
	YWRITE4(sc, YMF_MODE, mode & ~(YMF_MODE_ACTV|YMF_MODE_ACTV2));

	return 0;
}

int
ymf_halt_output(addr)
	void *addr;
{
	struct ymf_softc *sc = addr;

	DPRINTF(("ymf: ymf_halt_output\n"));
	if (sc->sc_play.intr) {
		sc->pbank[0]->status =
		sc->pbank[1]->status =
		sc->pbank[2]->status =
		sc->pbank[3]->status = 1;
		sc->pbank[0]->pgbase =
		sc->pbank[1]->pgbase =
		sc->pbank[2]->pgbase =
		sc->pbank[3]->pgbase = 0;
		sc->sc_play.intr = 0;
	}

	return 0;
}

int
ymf_halt_input(addr)
	void *addr;
{
	struct ymf_softc *sc = addr;

	DPRINTF(("ymf: ymf_halt_input\n"));
	sc->sc_rec.intr = NULL;
	if (sc->sc_rec.intr) {
		sc->rbank[0].pgbase =
		sc->rbank[1].pgbase =
		sc->rbank[2].pgbase =
		sc->rbank[3].pgbase = 0;
		sc->sc_rec.intr = 0;
	}

	return 0;
}

int
ymf_getdev(addr, retp)
	void *addr;
	struct audio_device *retp;
{
	*retp = ymf_device;

	return 0;
}

int
ymf_mixer_set_port(addr, cp)
	void *addr;
	mixer_ctrl_t *cp;
{
	struct ymf_softc *sc = addr;

	return (sc->sc_codec[0].codec_if->vtbl->mixer_set_port(
	    sc->sc_codec[0].codec_if, cp));
}

int
ymf_mixer_get_port(addr, cp)
	void *addr;
	mixer_ctrl_t *cp;
{
	struct ymf_softc *sc = addr;

	return (sc->sc_codec[0].codec_if->vtbl->mixer_get_port(
	    sc->sc_codec[0].codec_if, cp));
}

int
ymf_query_devinfo(addr, dip)
	void *addr;
	mixer_devinfo_t *dip;
{
	struct ymf_softc *sc = addr;

	return (sc->sc_codec[0].codec_if->vtbl->query_devinfo(
	    sc->sc_codec[0].codec_if, dip));
}

int
ymf_get_portnum_by_name(sc, class, device, qualifier)
	struct ymf_softc *sc;
	char *class, *device, *qualifier;
{
	return (sc->sc_codec[0].codec_if->vtbl->get_portnum_by_name(
	    sc->sc_codec[0].codec_if, class, device, qualifier));
}

void *
ymf_malloc(addr, direction, size, pool, flags)
	void *addr;
	int direction;
	size_t size;
	int pool, flags;
{
	struct ymf_softc *sc = addr;
	struct ymf_dma *p;
	int error;

	p = malloc(sizeof(*p), pool, flags);
	if (!p)
		return (0);
	error = ymf_allocmem(sc, size, 16, p);
	if (error) {
		free(p, pool);
		return (0);
	}
	p->next = sc->sc_dmas;
	sc->sc_dmas = p;
	return (KERNADDR(p));
}

void
ymf_free(addr, ptr, pool)
	void *addr;
	void *ptr;
	int pool;
{
	struct ymf_softc *sc = addr;
	struct ymf_dma **pp, *p;

	for (pp = &sc->sc_dmas; (p = *pp) != NULL; pp = &p->next) {
		if (KERNADDR(p) == ptr) {
			ymf_freemem(sc, p);
			*pp = p->next;
			free(p, pool);
			return;
		}
	}
}

static struct ymf_dma *
ymf_find_dma(sc, addr)
	struct ymf_softc *sc;
	void *addr;
{
	struct ymf_dma *p;

	for (p = sc->sc_dmas; p && KERNADDR(p) != addr; p = p->next)
		;

	return p;
}

size_t
ymf_round_buffersize(addr, direction, size)
	void *addr;
	int direction;
	size_t size;
{
	return (size);
}

paddr_t
ymf_mappage(addr, mem, off, prot)
	void *addr;
	void *mem;
	off_t off;
	int prot;
{
	struct ymf_softc *sc = addr;
	struct ymf_dma *p;

	if (off < 0)
		return (-1);
	p = ymf_find_dma(sc, mem);
	if (!p)
		return (-1);
	return (bus_dmamem_mmap(sc->sc_dmatag, p->segs, p->nsegs, 
				off, prot, BUS_DMA_WAITOK));
}

int
ymf_get_props(addr)
	void *addr;
{
	return (AUDIO_PROP_MMAP | AUDIO_PROP_INDEPENDENT | 
		AUDIO_PROP_FULLDUPLEX);
}
